
require("matrix")
require("shaders")
require("object")


local winW, winH = love.graphics.getDimensions()


local fov = math.rad(90)
local nearClip = 0.001
local farClip = 10000

local ambientLight = 0.6
local diffuseLight = 0.4
local lightSourceDir = Vec3(0.5,1,0.1):normalize()
local skyColor = {0.7, 0.9, 1, 1}


local iterations = 3
local maxPortals = 2

local depthCanvasOpt = {
	format = "depth32fstencil8",
	readable = true
}

local function buildCanvasTree(iter, lastPortal)
	local res = {}
	res.canvas = love.graphics.newCanvas(winW, winH)
	res.depth = love.graphics.newCanvas(winW, winH, depthCanvasOpt)
	if iter < iterations then
		for i = 1, maxPortals do
			res[i] = buildCanvasTree(iter + 1, i)
		end
	end
	return res
end

local canvases = buildCanvasTree(0)

function resize3d(newWinW, newWinH)
	winW, winH = newWinW, newWinH
	canvases = buildCanvasTree(0)
end


local function portalView(camView, portal)
	local rotate = Mat4.fromAngleAxis(math.pi, Vec3.e_y)
	local view = camView * portal:localToGlobal() * rotate * portal.dest:globalToLocal()
	local planeNormal = camView:withoutTranslation() * portal:localToGlobal():withoutTranslation() * (-Vec3.e_z)
	local planePoint = camView * portal.pos
	local proj = Mat4.perspective(fov, winW/winH, nearClip, farClip*10):clipPerspective(planePoint, planeNormal)
	return proj, view
end


local function renderView(objects, portals, iter, view, projMat, viewMat)
	love.graphics.setCanvas({ view.canvas, depthstencil = view.depth })
	love.graphics.clear(skyColor)
	
	shader:send("proj_view", projMat * viewMat)
	
	for i, object in pairs(objects) do
		local render =
			    isInstance(object, PhysicalObject)
			and object.visible
			and (not object.firstPerson or iter ~= 0)
		if render then
			local parts = object:getModel()
			for j, part in ipairs(parts) do
				shader:send("model", part.modelMat)
				shader:send("portal_surface", not not part.portal)
				
				love.graphics.setWireframe(not not part.model.wireframe)
				love.graphics.setMeshCullMode(part.model.culling and "back" or "none")
				love.graphics.setColor(part.model.color)
				
				if iter < iterations then
					if part.portal then
						local portalIdx = portals.fromValue[object]
						love.graphics.stencil(function()
							love.graphics.setColorMask()
							love.graphics.draw(part.model.mesh, -winW/2, -winH/2)
						end, "replace", portalIdx, true)
					else
						love.graphics.stencil(function()
							love.graphics.setColorMask()
							love.graphics.draw(part.model.mesh, -winW/2, -winH/2)
						end, "replace", 0, true)
					end
				else
					love.graphics.draw(part.model.mesh, -winW/2, -winH/2)
				end
			end
		end
	end
	
	if iter < iterations then
		for i, portal in ipairs(portals.fromIndex) do
			local portalProj, portalView = portalView(viewMat, portal)
			renderView(objects, portals, iter+1, view[i], portalProj, portalView)
		end
		love.graphics.setCanvas({ view.canvas, depthstencil = view.depth })
		love.graphics.setShader()
		love.graphics.setColor(1,1,1)
		for i = 1, #portals.fromIndex do
			love.graphics.setStencilTest("equal", i)
			love.graphics.draw(view[i].canvas)
		end
		love.graphics.setStencilTest()
		love.graphics.setShader(shader)
	end
end

function render(objects, camera)
	local portals = {
		fromIndex = {},
		fromValue = {}
	}
	local portalIndex = {}
	for i, object in ipairs(objects) do
		if isInstance(object, Portal) then
			table.insert(portals.fromIndex, object)
			portals.fromValue[object] = #portals.fromIndex
		end
	end
	
	love.graphics.setDepthMode("lequal", true)
	love.graphics.setStencilTest()
	love.graphics.setShader(shader)
	shader:send("ambient_light", ambientLight)
	shader:send("diffuse_light", diffuseLight)
	shader:send("light_source_dir", lightSourceDir)
	shader:send("sky_color", skyColor)
	
	local proj = Mat4.perspective(fov, winW/winH, nearClip, farClip)
	renderView(objects, portals, 0, canvases, proj, camera:globalToLocal())
	
	love.graphics.setShader()
	love.graphics.setStencilTest()
	love.graphics.setCanvas()
	love.graphics.setColor(1,1,1)
	love.graphics.draw(canvases.canvas, winW/2, winH/2, 0, 1, -1, winW/2, winH/2)
	
	invertShader:send("bg_img", canvases.canvas)
	love.graphics.setShader(invertShader)
	
	local crossSize = 10
	love.graphics.setLineWidth(2)
	love.graphics.setColor(1,1,1)
	love.graphics.line(winW/2-crossSize, winH/2, winW/2+crossSize, winH/2)
	love.graphics.line(winW/2, winH/2-crossSize, winW/2, winH/2+crossSize)
	
	love.graphics.setShader()
end
