local function getDefault(classMt, key)
	if classMt.instanceDef[key] ~= nil then
		return classMt.instanceDef[key]
	elseif classMt.superMt ~= nil then
		return getDefault(classMt.superMt, key)
	end
end

local function getIndexer(classMt)
	if classMt.indexer ~= nil then
		return classMt.indexer
	elseif classMt.superMt ~= nil then
		return getIndexer(classMt.superMt)
	end
end

local function getAssigner(classMt)
	if classMt.assigner ~= nil then
		return classMt.assigner
	elseif classMt.superMt ~= nil then
		return getAssigner(classMt.superMt)
	end
end

local function indexInstance(obj, key)
	local classMt = getmetatable(obj).classMt
	if type(key) == "string" then
		return getDefault(classMt, key)
	else
		local indexer = getIndexer(classMt)
		if type(indexer) == "function" then
			return indexer(obj, key)
		elseif type(indexer) == "table" then
			return indexer[key]
		end
	end
end

local function assignInstance(obj, key, val)
	local classMt = getmetatable(obj).classMt
	if type(key) == "string" then
		rawset(obj, key, val)
	else
		local assigner = getAssigner(classMt)
		if type(assigner) == "function" then
			assigner(obj, key, val)
		elseif type(assigner) == "table" then
			assigner[key] = val
		elseif assigner == nil then
			error("Trying to assign non-string property on object without __newindex")
		else
			error("Invalid __newindex metaproperty on class")
		end
	end
end

local function makeInstance(class, ...)
	local classMt = getmetatable(class)
	assert(class.init ~= nil, "Trying to instantiate abstract class (has no 'init' method)")
	local obj = setmetatable({}, classMt.instanceMt)
	class.init(obj, ...)
	return obj
end

local function defaultToString(obj, type)
	local mt = getmetatable(obj)
	setmetatable(obj, nil)
	local s = tostring(obj)
	setmetatable(obj, mt)
	return s:gsub("table", type)
end

Object = setmetatable({}, {
	instanceMt = {
		__index = indexInstance,
		__newindex = assignInstance,
		__tostring = function(obj)
			return defaultToString(obj, "object")
		end
	},
	instanceDef = {},
	__call = makeInstance
})


function class(desc)
	local classMt = {
		instanceMt = {},
		instanceDef = {},
		superMt = getmetatable(Object),
		__call = makeInstance,
		__tostring = function(class)
			return defaultToString(class, "class")
		end
	}
	classMt.instanceMt.classMt = classMt
	local class = setmetatable({}, classMt)
	for k,v in pairs(desc) do
		if k == "super" then
			classMt.superMt = getmetatable(desc.super)
		elseif k == "__index" then
			classMt.indexer = desc.__index
		elseif k == "__newindex" then
			classMt.assigner = desc.__newindex
		elseif type(k) == "string" and k:sub(1,1) == "_" then
			if k:sub(2,2) == "_" then
				classMt.instanceMt[k] = v
			else
				classMt.instanceDef[k:sub(2,-1)] = v
			end
		else
			class[k] = v
		end
	end
	-- we have to copy instanceMt from superclass to child class, because metatables are raw-accessed.
	for k, v in pairs(classMt.superMt.instanceMt) do
		if classMt.instanceMt[k] == nil then
			classMt.instanceMt[k] = v
		end
	end
	return class
end

local function isChildClass(class1Mt, class2Mt)
	return class1Mt == class2Mt or (class1Mt.superMt ~= nil and isChildClass(class1Mt.superMt, class2Mt))
end

function isInstance(obj, class)
	if type(obj) ~= "table" then return false end
	local objMt = getmetatable(obj)
	return objMt ~= nil and isChildClass(objMt.classMt, getmetatable(class))
end