
require("class")

local Vec4

Vec3 = class({
	init = function(self, x, y, z)
		x, y, z = x or 0, y or 0, z or 0
		self[1], self[2], self[3] = x, y, z
	end,
	__newindex = function(self, key, val) rawset(self, key, val) end,
	__add = function(a, b)
		return Vec3(a[1]+b[1], a[2]+b[2], a[3]+b[3])
	end,
	__sub = function(a, b)
		return Vec3(a[1]-b[1], a[2]-b[2], a[3]-b[3])
	end,
	__unm = function(self)
		return Vec3(-self[1], -self[2], -self[3])
	end,
	__mul = function(self, s)
		if type(self) == "number" then self, s = s, self end
		assert(type(s) == "number", "Can't multiply Vec3 by "..type(s))
		return Vec3(self[1]*s, self[2]*s, self[3]*s)
	end,
	__div = function(self, s)
		assert(type(s) == "number", "Can't divide Vec3 by "..type(s))
		assert(s ~= 0, "Can't divide Vec3 by 0")
		return Vec3(self[1]/s, self[2]/s, self[3]/s)
	end,
	_cross = function(a, b)
		return Vec3(a[2]*b[3]-a[3]*b[2], a[3]*b[1]-a[1]*b[3], a[1]*b[2]-a[2]*b[1])
	end,
	_dot = function(a, b)
		return a[1]*b[1] + a[2]*b[2] + a[3]*b[3]
	end,
	_len = function(self)
		return math.sqrt(self[1]^2 + self[2]^2 + self[3]^2)
	end,
	_normalize = function(self)
		local l = self:len()
		return l == 0 and self or self / l
	end,
	__tostring = function(self)
		return table.concat({"(", self[1], "; ", self[2], "; ", self[3], ")"})
	end,
	_toVec4 = function(self)
		return Vec4(self[1], self[2], self[3], 1)
	end,
})

Vec3.e_x = Vec3(1,0,0)
Vec3.e_y = Vec3(0,1,0)
Vec3.e_z = Vec3(0,0,1)

Vec4 = class({
	init = function(self, x, y, z, w)
		x, y, z, w = x or 0, y or 0, z or 0, w or 0
		self[1], self[2], self[3], self[4] = x, y, z, w
	end,
	__newindex = function(self, key, val) rawset(self, key, val) end,
	__add = function(a, b)
		return Vec4(a[1]+b[1], a[2]+b[2], a[3]+b[3], a[4]+b[4])
	end,
	__sub = function(a, b)
		return Vec4(a[1]-b[1], a[2]-b[2], a[3]-b[3], a[4]-b[4])
	end,
	__unm = function(self)
		return Vec4(-self[1], -self[2], -self[3], -self[4])
	end,
	__mul = function(self, s)
		if type(self) == "number" then self, s = s, self end
		assert(type(s) == "number", "Can't multiply Vec4 by "..type(s))
		return Vec4(self[1]*s, self[2]*s, self[3]*s, self[4]*s)
	end,
	__div = function(self, s)
		assert(type(s) == "number", "Can't divide Vec4 by "..type(s))
		assert(s ~= 0, "Can't divide Vec4 by 0")
		return Vec4(self[1]/s, self[2]/s, self[3]/s, self[4]/s)
	end,
	_dot = function(a, b)
		return a[1]*b[1] + a[2]*b[2] + a[3]*b[3] + a[4]*b[4]
	end,
	__tostring = function(self)
		return table.concat({"(", self[1], "; ", self[2], "; ", self[3], "; ", self[4], ")"})
	end,
	_toVec3 = function(self)
		return Vec3(self[1], self[2], self[3])
	end,
})

local function mat4Idx(row, col)
	return col + (row-1)*4
end

local function sign(x)
	return (x > 0) and 1 or ((x < 0) and -1 or 0)
end

-- 4x4 matrices, stored in row-major order
-- (as is compatible with LÖVE >= 11.0)
Mat4 = class({
	init = function(self, data)
		data = data or {}
		for i = 1, 16 do
			self[i] = data[i] or 0
		end
	end,
	__newindex = function(self, key, val) rawset(self, key, val) end,
	__tostring = function(self)
		local parts = {"["}
		for r = 1, 4 do
			for c = 1, 4 do
				table.insert(parts, self[mat4Idx(r,c)])
				if c < 4 then table.insert(parts, " ") end
			end
			if r < 4 then table.insert(parts, " | ") end
		end
		table.insert(parts, "]")
		return table.concat(parts)
	end,
	__mul = function(a, b)
		if isInstance(b, Mat4) then
			local res = Mat4()
			for r = 1, 4 do
				for c = 1, 4 do
					local i = mat4Idx(r,c)
					for k = 1, 4 do
						res[i] = res[i] + a[mat4Idx(r,k)] * b[mat4Idx(k,c)]
					end
				end
			end
			return res
		elseif isInstance(b, Vec3) then
			local v = b:toVec4()
			local res4 = Vec4()
			for r = 1, 4 do
				for c = 1, 4 do
					res4[r] = res4[r] + a[mat4Idx(r,c)] * v[c]
				end
			end
			return res4:toVec3()
		end
	end,
	id = function()
		return Mat4({
			1, 0, 0, 0,
			0, 1, 0, 0,
			0, 0, 1, 0,
			0, 0, 0, 1
		})
	end,
	_clone = function(self)
		return Mat4(self)
	end,
	_transpose = function(self)
		local res = Mat4()
		for r = 1, 4 do
			for c = 1, 4 do
				res[mat4Idx(r,c)] = self[mat4Idx(c,r)]
			end
		end
		return res
	end,
	-- https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
	fromAngleAxis = function(angle, axis)
		local x, y, z = axis[1], axis[2], axis[3]
		local c, s = math.cos(angle), math.sin(angle)
		return Mat4({
			x*x*(1-c)+c,   x*y*(1-c)-z*s, x*z*(1-c)+y*s, 0,
			y*x*(1-c)+z*s, y*y*(1-c)+c,   y*z*(1-c)-x*s, 0,
			x*z*(1-c)-y*s, y*z*(1-c)+x*s, z*z*(1-c)+c,   0,
			0,             0,             0,             1
		})
	end,
	-- Rotation matrix from rotation vector (Vec3).
	-- Each component of 'rot' describes an intrinsic /!\ rotation around the corresponding axis.
	-- Rotations are performed in the order Y -> X -> Z. (useful for FPS games!)
	-- (Note: intrinsic Y-X-Z is equivalent to extrinsic Z-X-Y)
	fromAngles = function(rot)
		return Mat4.fromAngleAxis(rot[2], Vec3.e_y)
		     * Mat4.fromAngleAxis(rot[1], Vec3.e_x)
		     * Mat4.fromAngleAxis(rot[3], Vec3.e_z)
	end,
	-- Given a rotation matrix, returns the corresponding angles, as decribed above.
	-- (Derivation: just compute the generic rotation matrix and solve for the angles)
	_toAngles = function(self)
		x = math.asin(-self[7]) -- x is between -pi/2 and pi/2
		if self[7] == 1 or self[7] == -1 then
			-- pointing straight up/down, gimbal lock
			-- to resolve gimbal lock, we decide that z = 0.
			y = math.atan2(-self[7]*self[2], self[1])
			z = 0
		else
			y = math.atan2(self[3], self[11])
			z = math.atan2(self[5], self[6])
		end
		return Vec3(x, y, z)
	end,
	_withoutTranslation = function(self)
		local m = self:clone()
		m[4], m[8], m[12] = 0, 0, 0
		return m
	end,
	_toPositionAngles = function(self)
		local pos = Vec3(self[4], self[8], self[12])
		local orient = self:withoutTranslation():toAngles()
		return pos, orient
	end,
	translation = function(vec3)
		return Mat4({
			1, 0, 0, vec3[1],
			0, 1, 0, vec3[2],
			0, 0, 1, vec3[3],
			0, 0, 0, 1,
		})
	end,
	-- Transformation matrix, from local space (of object described by pos/orient) to global space.
	localToGlobal = function(pos, orient)
		return Mat4.translation(pos) * Mat4.fromAngles(orient)
	end,
	-- Transformation matrix, from global space to local space (of object described by pos/orient).
	-- We avoid inverting matrices generally by inverting the translation/rotation separately.
	globalToLocal = function(pos, orient)
		return Mat4.fromAngles(orient):transpose() * Mat4.translation(-pos)
	end,
	perspective = function(fov, aspect, near, far)
		local t = math.tan(fov / 2)
		local res = Mat4()
		res[1]  =  1 / (t * aspect)
		res[6]  =  1 / t
		res[11] = -(far + near) / (far - near)
		res[12] = -(2 * far * near) / (far - near)
		res[15] = -1
		return res
	end,
	_clipPerspective = function(self, point, normal)
		local C = Vec4(normal[1], normal[2], normal[3], -normal:dot(point))
		local m = self:clone()
		if C[4] > 0 then return m end
		local Q = Vec4()
		Q[1] = (sign(C[1]) + m[3]) / m[1]
		Q[2] = (sign(C[2]) + m[7]) / m[6]
		Q[3] = -1
		Q[4] = (1 + m[11]) / m[12]
		local row3 = C * (2 / C:dot(Q)) + Vec4(0,0,1,0)
		m[9], m[10], m[11], m[12] = row3[1], row3[2], row3[3], row3[4]
		return m
	end
})
