
require("class")
require("matrix")


local format = {
	{"VertexPosition", "float", 3},
	{"VertexTexCoord", "float", 2},
	{"VertexNormal", "float", 3},
}

local Model = class({
	init = function(self, verts, tris, texture, color)
		self.verts = verts
		for i=1, #self.verts do
			local v = self.verts[i]
			
			-- if not given uv coordinates, put in random ones
			if #v < 5 then
				v[4] = love.math.random()
				v[5] = love.math.random()
			end
			
			-- if not given normals, figure it out
			if #v < 8 then
				local p1 = Vec3(v[1], v[2], v[3])
				local sum = Vec3(0,0,0)
				-- find triangles containing i
				for j = 1, #tris, 3 do
					local k = tris[j] == i and 1 or (tris[j+1] == i and 2 or (tris[j+2] == i and 3 or 0))
					if k ~= 0 then
						local p2 = self:getVertex(tris[j + k % 3])
						local p3 = self:getVertex(tris[j + (k+1) % 3])
						
						sum = sum + (p3-p2):cross(p2-p1):normalize()
					end
				end
				local normal = sum:normalize()
				v[6], v[7], v[8] = normal[1], normal[2], normal[3]
			end
		end
		self.tris = tris
		
		self.mesh = love.graphics.newMesh(format, verts, "triangles")
		self.mesh:setVertexMap(tris)
		if texture then self.mesh:setTexture(texture) end
		self.color = color or {1,1,1}
		
		self.wireframe = false
		self.culling = true
	end,
	
	_getVertex = function(self, i)
		return Vec3(self.verts[i][1], self.verts[i][2], self.verts[i][3])
	end,
	_getTris = function(self)
		local tris = {}
		for i = 1, #self.tris, 3 do
			table.insert(tris, {
				self:getVertex(self.tris[i]), self:getVertex(self.tris[i+1]), self:getVertex(self.tris[i+2])
			})
		end
		return tris
	end
})


local function vertex(pos, u, v)
	return {pos[1], pos[2], pos[3], u, v}
end

local function genRectVertices(pos, side1, side2, uMax, vMax)
	local uMin, vMin = 0, 0
	if uMax == nil then uMin, vMin, vMax = nil, nil, nil end
	local s1, s2 = side1/2, side2/2
	local verts = {
		vertex(pos-s1-s2, uMin, vMin),
		vertex(pos+s1-s2, uMax, vMin),
		vertex(pos+s1+s2, uMax, vMax),
		vertex(pos-s1+s2, vMin, vMax)
	}
	local tris = {3,2,1, 1,4,3}
	return verts, tris
end

local function appendList(t1, t2)
	local n = #t1
	for i = 1, #t2 do
		t1[n+i] = t2[i]
	end
end

local function mergeVerts(v1, t1, v2, t2)
	local n1, n2 = #v1, #t1
	appendList(v1, v2)
	appendList(t1, t2)
	for i = n2+1, #t1 do
		t1[i] = n1 + t1[i]
	end
end


function newRect(side1, side2, tex, uSize, vSize)
	local verts, tris = genRectVertices(Vec3(), side1, side2, uSize, vSize)
	return Model(verts, tris, tex)
end

function newSquare(size, dir1, dir2, tex, uvSize)
	local verts, tris = genRectVertices(Vec3(), size*dir1, size*dir2, uvSize, uvSize)
	return Model(verts, tris, tex)
end


local portalW, portalH = 1.2, 2.2
local portalEdge = 0.1

function newPortalFrame()
	local w = Vec3(portalW, 0, 0)
	local h = Vec3(0, portalH, 0)
	local e = Vec3(0, 0, portalEdge)
	local verts, tris = genRectVertices(-e/2, h, w) -- Back face
	mergeVerts(verts, tris, genRectVertices(-w/2, e, h)) -- Left edge
	mergeVerts(verts, tris, genRectVertices( w/2, h, e)) -- Right edge
	mergeVerts(verts, tris, genRectVertices( h/2, e, w)) -- Top edge
	mergeVerts(verts, tris, genRectVertices(-h/2, w, e)) -- Bottom edge
	
	local front = genRectVertices(e/2, w, h) -- Outside front
	appendList(front, genRectVertices(e/2, w-2*Vec3(portalEdge, 0, 0), h-2*Vec3(0, portalEdge, 0))) -- Inside front
	mergeVerts(verts, tris, front, { 1,8,5, 1,5,2, 2,5,6, 2,6,3, 3,6,7, 3,7,4, 4,7,8, 4,8,1 })
	
	mergeVerts(verts, tris, genRectVertices(-w/2+Vec3(portalEdge, 0, 0), h-2*Vec3(0, portalEdge, 0), e)) -- Left inside edge
	mergeVerts(verts, tris, genRectVertices( w/2-Vec3(portalEdge, 0, 0), e, h-2*Vec3(0, portalEdge, 0))) -- Right inside edge
	mergeVerts(verts, tris, genRectVertices( h/2-Vec3(0, portalEdge, 0), w-2*Vec3(portalEdge, 0, 0), e)) -- Top inside edge
	mergeVerts(verts, tris, genRectVertices(-h/2+Vec3(0, portalEdge, 0), e, w-2*Vec3(portalEdge, 0, 0))) -- Bottom inside edge
	
	return Model(verts, tris, nil, Vec3(0,1,0))
end

function newPortalSurface()
	local portalVerts, portalTris = genRectVertices(Vec3(),
		Vec3(portalW - 2*portalEdge, 0, 0), Vec3(0, portalH - 2*portalEdge, 0), 0, 0)
	return Model(portalVerts, portalTris)
end


function newCuboid(pos, a, b, c, tex, uMax, vMax)
	local verts, tris = {}, {}
	local x, y, z = a*Vec3.e_x, b*Vec3.e_y, c*Vec3.e_z
	mergeVerts(verts, tris, genRectVertices(pos-x/2, z, y, uMax, vMax))
	mergeVerts(verts, tris, genRectVertices(pos+x/2, y, z, uMax, vMax))
	mergeVerts(verts, tris, genRectVertices(pos-z/2, y, x, uMax, vMax))
	mergeVerts(verts, tris, genRectVertices(pos+z/2, x, y, uMax, vMax))
	mergeVerts(verts, tris, genRectVertices(pos-y/2, x, z, uMax, vMax))
	mergeVerts(verts, tris, genRectVertices(pos+y/2, z, x, uMax, vMax))
	return Model(verts, tris, tex)
end

function newCube(size, tex)
	return newCuboid(Vec3(), size, size, size, tex, 1, 1)
end
