
require("class")
require("matrix")
require("model")
require("intersect")

GameObject = class({})

OrientableObject = class({
	super = GameObject,
	init = function(self, pos, orient)
		self.pos = pos or Vec3(0, 0, 0)
		self.orient = orient or Vec3(0, 0, 0)
	end,
	_localToGlobal = function(self)
		return Mat4.localToGlobal(self.pos, self.orient)
	end,
	_globalToLocal = function(self)
		return Mat4.globalToLocal(self.pos, self.orient)
	end,
	_setTransform = function(self, mat)
		self.pos, self.orient = mat:toPositionAngles()
	end,
})

PhysicalObject = class({
	super = OrientableObject,
	init = function(self, pos, orient, grabbable)
		OrientableObject.init(self, pos, orient)
		self.grabbable = grabbable or false
		self.visible = true
	end,
	_getModel = function(self) error("Not implemented") end,
	_getHitbox = function(self)
		local tris = {}
		local parts = self:getModel()
		for i, part in ipairs(parts) do
			local partTris = part.model:getTris()
			local trans = part.modelMat
			for j, tri in ipairs(partTris) do
				local a, b, c = tri[1], tri[2], tri[3]
				table.insert(tris, { trans * a, trans * b, trans * c })
			end
		end
		return tris
	end,
	_raycast = function(self, src, dir)
		local triangles = self:getHitbox()
		local hit, dist = nil, nil
		for i, tri in ipairs(triangles) do
			local a, b, c = tri[1], tri[2], tri[3]
			local hit2, dist2 = intersectRayTri(src, dir, {a,b,c})
			if hit2 and (hit == nil or dist2 < dist) then
				hit, dist = hit2, dist2
			end
		end
		return hit, dist
	end,
})

local playerHeight = 1.7
local headSize = 0.3
local eyeHeight = playerHeight - headSize/2
local bodyHeight = playerHeight - headSize
local playerWidth = 0.3

Player = class({
	super = PhysicalObject,
	init = function(self, pos, orient)
		PhysicalObject.init(self, pos, orient, false)
		self.head = newCuboid(Vec3(), headSize, headSize, headSize)
		self.body = newCuboid(Vec3(), playerWidth, bodyHeight, playerWidth)
		self.firstPerson = true
	end,
	_getCameraPos = function(self)
		return self.pos + Vec3(0, eyeHeight, 0)
	end,
	_getCamera = function(self)
		return OrientableObject(self:getCameraPos(), self.orient)
	end,
	_getModel = function(self)
		local bodyModelMat = Mat4.localToGlobal(self.pos + Vec3(0, bodyHeight / 2), Vec3(0, self.orient[2], 0))
		return {
			{ model = self.head, modelMat = self:getCamera():localToGlobal() },
			{ model = self.body, modelMat = bodyModelMat }
		}
	end,
	_getHitbox = function(self)
		return {}
	end,
})

Portal = class({
	super = PhysicalObject,
	init = function(self, pos, orient)
		PhysicalObject.init(self, pos, orient, false)
		self.portalFrame = newPortalFrame()
		self.portalSurface = newPortalSurface()
		self.dest = nil
	end,
	_getModel = function(self)
		return {
			{ model = self.portalFrame, modelMat = self:localToGlobal() },
			{ model = self.portalSurface, modelMat = self:localToGlobal(), portal = true },
		}
	end,
})

SimpleObject = class({
	super = PhysicalObject,
	init = function(self, model, pos, orient, grabbable)
		PhysicalObject.init(self, pos, orient, grabbable)
		self.model = model
	end,
	_getModel = function(self)
		return {{ model = self.model, modelMat = self:localToGlobal() }}
	end,
})

local function defaultGroundTexture()
	local tex = love.graphics.newImage("ground.png")
	tex:setFilter("nearest", "nearest")
	return tex
end

Ground = class({
	super = SimpleObject,
	init = function(self, size, texture, textureSize)
		texture = texture or defaultGroundTexture()
		textureSize = textureSize or 1
		texture:setWrap("repeat", "repeat")
		local model = newSquare(size, Vec3(0,0,1), Vec3(1,0,0), texture, size / textureSize)
		SimpleObject.init(self, model, Vec3(), Vec3(), false)
	end
})

Cube = class({
	super = SimpleObject,
	init = function(self, pos, orient, size, texture)
		pos = pos or Vec3()
		orient = orient or Vec3()
		size = size or 0.75
		texture = texture or love.graphics.newImage("trans.png")
		local model = newCube(size, texture)
		SimpleObject.init(self, model, pos, orient, true)
	end
})
